/* Debug Header */
#if DEBUG
# define IFDEBUG(x) do { if (debug) x; }while(0)
# define IFDEBUG5(x) do { if (debug) x; }while(0)
#else
# define IFDEBUG(x)
# define IFDEBUG5(x)
#endif
/* End Debug Header */


#if HAVE_CONFIG_H
# include <config.h>
#endif


#include <sys/file.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <stdarg.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include "dbconfig.h"
#include "extern.h"
#include "defines.h"

/* dbconf_struct *config;
 * dbcon_struct *db;
 */

/* main config file handler
 * parse_config():
 *
 * Accepts: argc, and argv
 * Returns: 1 for error, 0 for sucsess
 *
 * purpose: populate config structure.
 */

int parse_config(int argc, char **argv) {

	char conf_file[]=CONFIG_FILE_NAME;
	int fd, i;
	

	config=(dbconf_struct *)calloc(sizeof(dbconf_struct), 1);

	if (fd=lock_config(conf_file)) {
		
		IFDEBUG(print_debug(4, "Reading config file\n"));
		read_config(fd);

		IFDEBUG(print_debug(4, "Unlocking/closing config file\n"));
		release_config(fd);

		return 0;
	}
	return 1;
}

void release_config(int fd) {

	IFDEBUG(print_debug(3, "Closing Config File\n"));
	lockf(fd, F_ULOCK, 0);
	close(fd);

}

/* Lock config file
 * lock_config():
 *
 * accepts: filename
 * returns: file descriptor
 *
 * purpose: open and lock the config file
 */

int lock_config(char *filename) {

	int fd;

	IFDEBUG(print_debug(1, "Opening Config File\n"));
	fd=open(filename, O_RDONLY);
	if (errno) {
		fprintf(stderr, "%s: ", filename);
		perror("");
		return 0;
	}
	
	while (lockf(fd, F_TEST, 0)) {
		if (errno != EACCES) {
			perror("Error Locking Config file");
			exit(1);
		}
		fprintf(stderr, "Waiting for config file lock!\n");
		sleep(5);
	}

	IFDEBUG(print_debug(4, "Locking Config File For exclusive access\n"));
	lockf(fd, F_LOCK, 0);
	IFDEBUG(print_debug(4, "Got Config Lock\n"));

	return fd;
}

/* read config
 * read_config()
 * accepts: open and locked filedescriptor for config file
 * returns: nothing
 *
 * purpose: read in configfile and set options/globals
 */
void read_config(int fd) {
	
	int i;
	char *p, *line;
	char *buf;

	print_debug(4, "Reading Config File into Buffer\n");
	config->config=read_file(fd);

	buf=malloc(FILE_BUF);
	line=config->config;
	while (line && *line) {
		memset(buf, '\0', FILE_BUF);
		p=(char *)strchr(line, '\n');
		if (p) {
			strncpy(buf, line, (p-line)<FILE_BUF?(size_t)(p-line):(size_t)FILE_BUF-1);
			parse_line(buf);
			p++;
		}
		line=p;
	}
	free(buf);
}

void parse_line(char *buf) {

	char *p;
	int i;

	while (*buf && (*buf == '\r' || *buf == '\n' || *buf == '\t')) buf++; /* Skip to first char */

	if (*buf == '#') return;

	p=buf;
	while (*p) {
		*p=tolower(*p); /* kill case */
		p++;
	}
	p=buf;

	IFDEBUG(print_debug(5, "Buf: \"%s\"\n", buf));
	i=getoption(buf);
	IFDEBUG(print_debug(5, "getoption() returned: %i\n", i));
	if (!i) return;
	buf=(char *)strchr(buf, '=');
	buf++;
	switch(i) {
		case 1:
			config->dbhome=(char *)calloc(sizeof(char), strlen(buf)+1);
			strcpy(config->dbhome, buf);
			break;
		case 2:
			config->dbport=atoi(buf);
			break;
		case 3:
			config->dbmode=getmode(buf);
			break;
		case 4:
			config->dbtype=gettype(buf);
			break;
		case 5:
			config->http=(char *)calloc(sizeof(char), strlen(buf)+1);
			strcpy(config->http, buf);
			break;
		case 6:
			config->rdns=(char *)calloc(sizeof(char), strlen(buf)+1);
			strcpy(config->rdns, buf);
			break;
		case 7:
			config->smtp=(char *)calloc(sizeof(char), strlen(buf)+1);
			strcpy(config->smtp, buf);
			break;
		case 8:
			config->sock=(char *)calloc(sizeof(char), strlen(buf)+1);
			strcpy(config->sock, buf);
			break;
		case 9:
			config->trap=(char *)calloc(sizeof(char), strlen(buf)+1);
			strcpy(config->trap, buf);
			break;
		case 10:
			config->oklist=(char *)calloc(sizeof(char), strlen(buf)+1);
			strcpy(config->oklist, buf);
			break;
		case 11:
			config->white=(char *)calloc(sizeof(char), strlen(buf)+1);
			strcpy(config->white, buf);
			break;
		case 12:
			config->mode=getmode(buf);
			break;
		case 13:
			config->threads=getbool(buf);
			break;
		case 14:
			debug=atoi(buf);
			break;
		case 15:
			config->listenport=atoi(buf);
			break;
		case 16:
			config->listenaddr=(char *)calloc(sizeof(char), strlen(buf)+1);
			strcpy(config->listenaddr, buf);
			break;
		case 17:
			config->logfile=(char *)calloc(sizeof(char), strlen(buf)+1);
			strcpy(config->logfile, buf);
			break;
		case 18:
			config->errfile=(char *)calloc(sizeof(char), strlen(buf)+1);
			strcpy(config->errfile, buf);
			break;
		case 19:
			config->lockfile=(char *)calloc(sizeof(char), strlen(buf)+1);
			strcpy(config->lockfile, buf);
			break;
		case 20:
			get_db((char *)strdup(buf)); /* strdup() cause we need to mod the contents. get_db() will call free() */
			break;
		case 21:
			db->unix_socket=(char *)strdup(buf);
			break;
		default:
			break;
		}
}

/* getbool():
 * accepts: pointer to string
 * returns: 1 (for true) or 0 (for false)
 *
 * purpose: will return 1 (one) when matching '1', 'on', 'yes' or 'true'
 *          returns 0 (zero) in all other cases.
 *
 */

int getbool(char *buf) {

	char *state[] = { "1",
	                  "on",
	                  "true",
	                  "yes",
	                  "" };
	int i;

	if (!buf || !*buf) return 0;

	for(i=0;;i++) {
		if (!*state[i]) goto false;
		if (!strcmp(buf, state[i])) goto true;
	}
false:
	IFDEBUG(print_debug(4, "getbool():false: '%s' returning: 0\n", buf));
	return 0;
true:
	IFDEBUG(print_debug(4, "getbool():true: '%s' returning: 1\n", buf));
	return 1;
}

int gettype(char *buf) {

	/* Types:
	 * 0 - Error
	 * 1 - local
	 * 2 - mysql
	 * 3 - pgsql
	 * 4 - msql
	 * 5 - iodbc
	 * 6 - unixodbc
	 * 7 - db2
	 * 8 - solid
	 * 9 - openlink
	 * 10 - easysoft
	 * 11 - sapdb
	 * 12 - ibase
	 * 13 - ctlib
	 * 14 - oracle8
	 * 15 - oracle7
	 */

	char *types[] = { "",
										"local",
	                  "mysql",
	                  "pgsql",
	                  "msql",
										"iodbc",
										"unixodbc",
										"db2",
										"solid",
										"openlink",
										"easysoft",
										"sapdb",
										"ibase",
										"ctlib",
										"oracle8",
										"oracle7",
	                  "" };
	int i;

	IFDEBUG(print_debug(4, "gettype(): %s\n", buf));
	if (!buf || !*buf) return 0;

	for(i=1;;i++) {
		if (!strlen(types[i])) return 0;
		if (!strcmp(buf, types[i])) return i;
	}

}

void get_db(char *buf) {

  char *t, *s, *e, *p;
	int bad=0;

	db=calloc(sizeof(dbcon_struct), 1);
	t=buf;

	s=NULL;

	IFDEBUG(print_debug(5, "Parsing DB: "));
	p=strchr(t, ':');
	if (p && !db->type) {
		*p='\0';
		db->type=gettype(t);
		t=++p;
		t++;
		t++;
	} else bad++;
	p=strchr(t, ':');
	if (p && !db->user) {
		*p='\0';
		db->user=strdup(t);
		t=++p;
	} else bad++;
	p=strchr(t, '@');
	if (p && !db->pass) {
		*p='\0';
		db->pass=strdup(t);
		t=++p;
	} else bad++;
	p=strchr(t, '/');
	if (p && !db->host) {
		*p='\0';
		if (s=strchr(t, ':')) {
			*s++='\0';
		}
		db->host=strdup(t);
		if (s) {
			db->port=atoi(s);
			s=NULL;
		}
		t=++p;
	} else bad++;
	if (!p && !db->name) p=t; /* sanity check */
	
	/* the next line is left over from another project, but makes for another safety check */
	while (*p && *p != '/' && *p != ' ' && *p != '\t' && *p != '\r' && *p != '\n') p++;
	*p='\0';
	if (*t) db->name=strdup(t);
	else bad++;
	if (bad) {
		IFDEBUG(print_debug(5, "Bad Parse, ignoring....\n"));
		if (db->type) db->type=0;
		if (db->user) free(db->user);
		if (db->pass) free(db->pass);
		if (db->host) free(db->host);
		if (db->name) free(db->name);
	} else IFDEBUG(print_debug(5, "type=%i, username='%s', password is not shown, host='%s', dbname='%s'\n", db->type, db->user, db->host, db->name));
	free(buf);
}

int getmode(char *buf) {

	/* types:
	 * 0 - standalone
	 * 1 - daemon
	 * 2 - plugin (Netscape)
	 * 3 - plugin (iMS)
	 */

	char *types[] = { "local",
	                  "daemon",
			              "nms",
			              "ims",
	                  "" };
	int i;

	IFDEBUG(print_debug(4, "getmode(): %s\n", buf));
	if (!buf || !*buf) return 0;

	for(i=0;;i++) {
		if (!strlen(types[i])) return 0;
		if (!strcmp(buf, types[i])) return i;
	}
	return 0;
}

int getoption(char *buf) {

	int i;
	char *p;

	char *options[] = { "db_home",		/* 1 */
											"db_port",		/* 2 */
											"db_mode",		/* 3 */
											"db_type",		/* 4 */
											"file_http",		/* 5 */
											"file_rdns",		/* 6 */
											"file_smtp",		/* 7 */
											"file_socks",		/* 8 */
											"file_spamtrap",		/* 9 */
											"local_db",		/* 10 */
											"file_whitelist", /* 11 */
											"mode",		/* 12 */
											"threads",		/* 13 */
											"debug",		/* 14 */
											"dblisten_port",		/* 15 */
											"dblisten_addr",		/* 16 */
											"db_logfile",		/* 17 */
											"db_errfile",		/* 18 */
											"lockfile",		/* 19 */
											"db_url",					/* 20 */
											"dbsocket",		/* 21 */
											"" };

	if (!buf || !*buf) return 0;
	p=(char *)strchr(buf, '=');
	if (!p) return 0;
	for (i=0;*options[i];i++) {
		if (!strlen(options[i])) return 0;
		// printf("Attempting to match the first %i chars of: %s & %s\n", (int)(p-buf), options[i], buf);
		if (!strncmp(options[i], buf, (size_t)(p-buf))) return i+1; /* Add 1 to the return! */
	}
	return 0;
}

/* close config */


/* read_file()
 * accepts: opend filedescriptor
 * returns: pointer
 *
 * purpose: allocate memory and read-in contents of file
 */
char *read_file(int fd) {

	int l, t;
	char *p, *buf;

	p=NULL;
	t=0;

	buf=malloc(FILE_BUF);
	
	l=1;
	while (l) {
		memset(buf, '\0', FILE_BUF);
		l=read(fd, buf, FILE_BUF-1);
		if (l==-1) {
			perror("read_file()");
			exit (1);
		}
		IFDEBUG(print_debug(5, "Read %i chars into the buffer\n", l));
		if (!l) break;
		t+=l;
		if (p) {
			p=realloc(p, t);
			strcat(p, buf);
		} else {
			p=strdup(buf);
		}
	}
	free(buf);
	return p;
}

/*
void print_config(void) {

	int i;

	print_debug(5,"Config File Reads:\n%s\n", config->config);
	print_debug(1,"Add open HTTP Proxies: %s\n", config->add_http?"true":"false");
	print_debug(1,"Add open SMTP Relays: %s\n", config->add_smtp?"true":"false");
	print_debug(1,"Add open Socks Proxies: %s\n", config->add_sock?"true":"false");
	print_debug(1,"Add hosts with no reverse DNS: %s\n", config->add_rdns?"true":"false");
	print_debug(1,"Add hosts that try to use the SPAM trap: %s\n", config->add_trap?"true":"false");
	print_debug(1,"File listing open HTTP Proxies: %s\n", config->file->http);
	print_debug(1,"File listing open SMTP Relays: %s\n", config->file->smtp);
	print_debug(1,"File listing open Socks Proxies: %s\n", config->file->sock);
	print_debug(1,"File listing hosts with no reverse DNS: %s\n", config->file->rdns);
	print_debug(1,"File listing hosts that tried to use the SPAM trap: %s\n", config->file->trap);
	print_debug(1,"Reject open HTTP Proxies: %s\n", config->add_http?"true":"false");
	print_debug(1,"Reject open Socks Proxies: %s\n", config->reject_sock?"true":"false");
	print_debug(1,"Reject open SMTP Relays: %s\n", config->reject_smtp?"true":"false");
	print_debug(1,"Reject hosts with no reverse DNS: %s\n", config->reject_rdns?"true":"false");
	print_debug(1,"Reject mail that is passed through open servers (socks/http proxies/smtp relays): %s\n", config->reject_body?"true":"false");
	print_debug(1,"Reject mailers that have used the spamtrap: %s\n", config->reject_trap?"true":"false");

	print_debug(1,"Database host: %s\n", config->dbhome);
	print_debug(1,"Database port: %i\n", config->dbport);
	print_debug(1,"Database mode: %i\n", config->dbmode);
	print_debug(1,"Database type: %i\n", config->dbtype);

	print_debug(1,"%i Ports to test for open servers: \n", config->ports);
	for (i=0;i<config->ports;i++) {
		print_debug(1, "%i [%s]\n", config->port[i]->port, config->port[i]->type?config->port[i]->type==SMTP?"SMTP":config->port[i]->type==HTTP?"HTTP":"SOCKS":"IDLE");
	}

	if (config->spamtrap) {
		for (i=0;config->spamtrap[i];i++) {
			print_debug(1, "Spam trap email address(s): %s\n", config->spamtrap[i]);
		}

	}

}
*/

void cleanup(void) {

	char *p;
	int i;

	cleanup_threads();
	if (config->http) free(config->http);
	if (config->smtp) free(config->smtp);
	if (config->sock) free(config->sock);
	if (config->rdns) free(config->rdns);
	if (config->trap) free(config->trap);
	if (config->oklist) free(config->oklist);

	if (config->dbhome) free(config->dbhome);

	if (config->config) free(config->config);

	if (config->listenaddr) free(config->listenaddr);

	free(config);

}
